#!/usr/bin/env python3
"""
keyrings_artifacts/artifacts.py
---------------------------

Azure Artifacts keyring backend.
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

import os
import warnings
from typing import TYPE_CHECKING
from urllib.parse import urlsplit

import keyring.credentials

from .provider import CredentialProvider


class ArtifactsKeyringBackend(keyring.backend.KeyringBackend):
    """
    Azure Artifacts keyring backend.

    This backend provides authentication for publishing or consuming Python packages to or from
    Azure Artifacts feeds within [Azure DevOps](https://azure.com/devops). Credentials are retrieved
    from environment variables, managed identity, the Azure CLI, shared token cache, or the device
    code flow.
    """

    SUPPORTED_NETLOC = ("pkgs.dev.azure.com", "pkgs.visualstudio.com", "pkgs.codedev.ms", "pkgs.vsts.me")
    _PROVIDER = CredentialProvider

    __LOCAL_BACKEND = None

    priority = 9.9

    @property
    def _LOCAL_BACKEND(self):
        """
        Get the local keyring backend.
        """
        # if we fall back to the local backend, we don't want to loop infinitely
        # so we set the priority to -1 to prevent it from being selected again
        if self.__LOCAL_BACKEND is not None:
            return self.__LOCAL_BACKEND

        _self = next(
            kr for kr in keyring.backend.get_all_keyring() if kr.__class__.__name__ == "ArtifactsKeyringBackend"
        )
        _self.priority = -1
        self.__LOCAL_BACKEND = keyring.get_keyring()
        return self.__LOCAL_BACKEND

    def __init__(self):
        self._cache = {}

    def get_credential(self, service: str, username: str | None) -> keyring.credentials.SimpleCredential | None:
        """
        Retrieve credentials for a given service.

        Parameters
        ----------
        service : str
            The service URL.
        username : Optional[str]
            The username for the service.

        Returns
        -------
        Optional[keyring.credentials.SimpleCredential]
            The credentials for the service, or None if not found.
        """
        try:
            parsed = urlsplit(service)
        except Exception as exc:
            warnings.warn(str(exc))  # noqa: B028
            return None

        netloc = parsed.netloc.rpartition("@")[-1]

        if netloc is None or not netloc.endswith(self.SUPPORTED_NETLOC):
            return None

        provider = self._PROVIDER()
        use_bearer_token = os.getenv(provider._USE_BEARER_TOKEN_VAR_NAME, "False").lower() == "true"

        if not use_bearer_token and username:
            # Check if credentials are already stored in the local keyring
            stored_password = self._LOCAL_BACKEND.get_password(service, username)

            if stored_password and provider._can_authenticate(service, (self._PROVIDER.username, stored_password)):
                return keyring.credentials.SimpleCredential(username, stored_password)

        # else or if the stored password is not valid, try to retrieve the credentials from the provider
        username, password = provider.get_credentials(service)

        if username and password:
            if not use_bearer_token:
                # Store the retrieved username and PAT in the local keyring
                self._LOCAL_BACKEND.set_password(service, username, password)
                self._cache[service, username] = password
            return keyring.credentials.SimpleCredential(username, password)

    def get_password(self, service: str, username: str) -> str | None:
        """
        Retrieve the password for a given service and username.

        Parameters
        ----------
        service : str
            The service URL.
        username : str
            The username for the service.

        Returns
        -------
        Optional[str]
            The password for the service, or None if not found.
        """
        password = self._cache.get((service, username), None)
        if password is not None:
            return password

        creds = self.get_credential(service, username)
        if creds and username == creds.username:
            return creds.password

        return None

    def set_password(self, service: str, username: str, password: str) -> None:
        """
        Set the password for a given service and username.

        Parameters
        ----------
        service : str
            The service URL.
        username : str
            The username for the service.
        password : str
            The password to set.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        # Defer setting a password to the next backend
        raise NotImplementedError()

    def delete_password(self, service: str, username: str) -> None:
        """
        Delete the password for a given service and username.

        Parameters
        ----------
        service : str
            The service URL.
        username : str
            The username for the service.

        Raises
        ------
        NotImplementedError
            This method is not implemented.
        """
        self._LOCAL_BACKEND.delete_password(service, username)


if TYPE_CHECKING:
    ArtifactsKeyringBackend()



#!/usr/bin/env python3
"""
artifacts-keyring version file
------------------------------
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

TYPE_CHECKING = False
if TYPE_CHECKING:
    VERSION_TUPLE = tuple[int | str, ...]

version: str
__version__: str
__version_tuple__: VERSION_TUPLE
version_tuple: VERSION_TUPLE

__version__ = "0.4.4"  # version string
vesion = __version__
__version_tuple__ = version_tuple = tuple(int(n) for n in __version__.split(".") if n.isdigit())



#!/usr/bin/env python3
"""
keyrings_artifacts/provider.py
---------------------------

This module implements the CredentialProvider class which is a Azure credential providers for
authenticating with Azure DevOps Artifacts.
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

import logging
import os
import re
from datetime import datetime, timedelta, timezone
from http import HTTPStatus

import requests
from azure.core.exceptions import ClientAuthenticationError
from requests.exceptions import HTTPError, RequestException

from .support import AzureCredentialWithDevicecode

# Set up logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

logging.getLogger("azure.identity._credentials.managed_identity").setLevel(logging.ERROR)
logging.getLogger("azure.identity._credentials.environment").setLevel(logging.ERROR)
logging.getLogger("azure.identity._credentials.vscode").setLevel(logging.ERROR)
logging.getLogger("azure.core.pipeline.policies.http_logging_policy").setLevel(logging.ERROR)


class CredentialProvider:
    """
    Credential provider for Azure DevOps Artifacts.

    This plugin provides credentials for Azure DevOps Artifacts.
    """

    # Non-PAT, bearer token mode
    _USE_BEARER_TOKEN_VAR_NAME = "KEYRINGS_ARTIFACTS_USE_BEARER_TOKEN"

    # ADO global variables
    _PAT_ENV_VAR = "AZURE_DEVOPS_EXT_PAT"
    _ADO_USERNAME_ENV_VAR = "AZURE_DEVOPS_USERNAME"
    _PAT_SCOPE_ENV_VAR = "AZURE_DEVOPS_PAT_SCOPE"
    _PAT_DURATION_ENV_VAR = "AZURE_DEVOPS_PAT_DURATION"

    # Azure DevOps application default scope
    # from https://github.com/microsoft/artifacts-credprovider/blob/cdc427e8236212b33041b4276961855b39bbe98d/CredentialProvider.Microsoft/CredentialProviders/Vsts/MSAL/MsalTokenProviderFactory.cs#L11
    _OAUTH_SCOPE = "499b84ac-1321-427f-aa17-267ca6975798/.default"

    # PAT global variables
    # PAT API route
    _TOKEN_API_ROUTE = "/_apis/tokens/pats?api-version=7.1-preview.1"
    # User-Agent for the request
    _USER_AGENT = "PyMSALCredentialProvider"
    # Display name for the PAT
    _DISPLAY_NAME = "Azure DevOps Artifacts Credential Provider"
    # PAT duration in days
    _DEFAULT_PAT_DURATION = 365
    # Default username for the credentials
    _DEFAULT_USERNAME = "VssSessionToken"
    # Default PAT scope
    _DEFAULT_VSTS_SCOPE = "vso.packaging_write"

    @property
    def username(self) -> str | None:
        """Get the username."""
        return os.getenv(self._ADO_USERNAME_ENV_VAR, self._DEFAULT_USERNAME)

    def _is_upload_endpoint(self, url: str) -> bool:
        """Check if the given URL is the upload endpoint."""
        return url.rstrip('/').endswith("pypi/upload")

    def _can_authenticate(self, url: str, auth: tuple[str, str] | None) -> bool:
        """Check if the given URL can be authenticated with the given credentials."""
        response = requests.get(url, auth=auth)

        return response.status_code < HTTPStatus.INTERNAL_SERVER_ERROR and response.status_code not in (
            HTTPStatus.UNAUTHORIZED,
            HTTPStatus.FORBIDDEN,
        )

    def _get_authorities(self, url: str) -> tuple[str, str]:
        """Send a GET to the url and parse the response header"""
        # send GET request
        response = requests.get(url)
        headers = response.headers

        # extract oauth authority and tenant_id
        match = re.search(r'Bearer authorization_uri=(https://[^/]+/)[^,]+', headers["WWW-Authenticate"])
        if match:
            bearer_authority = match.group(1)
            tenant_id = match.group(0).rsplit("/",1)[1]
        else:
            # the Azure DevOps endpoint seems to be linked to a personal Microsoft account
            # and only basic authentication or PAT is supported
            bearer_authority = ""
            tenant_id = ""
        return bearer_authority, tenant_id, headers["X-VSS-AuthorizationEndpoint"]

    def _get_bearer_token(
        self, authority: str, tenant_id: str, scope: str
    ) -> str:
        """
        Get the bearer token for the given URL.

        This method is used to get the bearer token for the given URL. The token is obtained
        using the AzureCredentialWithDevicecode class which tries to optain the token from:
        1) Environment variables
        2) Azure CLI
        3) Shared Token Cache
        4) Interactive Browser
        5) DeviceCode flow
        """
        try:
            token = (
                AzureCredentialWithDevicecode(
                    tenant_id=tenant_id,
                    authority=authority
                )
                .get_token(scope)
                .token
            )
        except ClientAuthenticationError as e:
            if "Azure Active Directory error" not in str(e):
                raise e
            # DefaultAzureCredential raises an exception when there is a token
            # found in the cache but the token has expired. In this case we catch the error and
            # initiate the an Interactive Browser flow if possible or fall back to the DeviceCode flow.
            logger.warning(f"Caught {e.__class__}: {e!s}! Falling back to Interactive Browser flow.")
            token = (
                AzureCredentialWithDevicecode(
                    authority=authority,
                    tenant_id=tenant_id,
                    with_az_cli=False
                )
                .get_token(scope)
                .token
            )
        return token

    def _exchange_bearer_for_pat(self, authority_endpoint: str, bearer_token: str) -> str:
        """Exchange the bearer token for a personal access token (PAT)."""
        try:
            # Build request headers
            request_headers = {
                "Authorization": f"Bearer {bearer_token}",
                "Content-Type": "application/json",
                "User-Agent": self._USER_AGENT,
            }

            # Build the request URL
            visual_studio_url = f"{authority_endpoint.rstrip('/')}/{self._TOKEN_API_ROUTE.lstrip('/')}"

            # Build the request payload
            _delta = timedelta(days=int(os.getenv(self._PAT_DURATION_ENV_VAR, self._DEFAULT_PAT_DURATION)))
            expiry = (datetime.now(timezone.utc) + _delta).strftime("%Y-%m-%dT%H:%M:%SZ")

            request_payload = {
                "displayName": self._DISPLAY_NAME,
                "scope": os.getenv(self._PAT_SCOPE_ENV_VAR, self._DEFAULT_VSTS_SCOPE),
                "validTo": expiry,
                "allOrgs": "false",
            }
            # Send request
            with requests.post(visual_studio_url, headers=request_headers, json=request_payload) as response:
                response.raise_for_status()  # Raise an HTTPError for bad responses

                # Return the PAT token
                return response.json()["patToken"]["token"]

        except HTTPError as http_err:
            print(f"HTTP error occurred: {http_err}")
            raise
        except RequestException as req_err:
            print(f"Request error occurred: {req_err}")
            raise
        except KeyError as key_err:
            print(f"Key error occurred: {key_err}")
            raise
        except Exception as err:
            print(f"An error occurred: {err}")
            raise

    def _get_credentials_from_credential_provider(self, url: str) -> tuple[str | None, str | None]:
        """Get the credentials from the credential provider."""

        # get username
        username = os.getenv(self._ADO_USERNAME_ENV_VAR, self._DEFAULT_USERNAME)
        # get authorities and tenant_id
        oauth_authority, tenant_id, vsts_authority = self._get_authorities(url)
        # get bearer token
        bearer_token = self._get_bearer_token(
            authority=oauth_authority,
            tenant_id=tenant_id,
            scope=self._OAUTH_SCOPE,
        )
        if os.getenv(self._USE_BEARER_TOKEN_VAR_NAME, "False").lower() == "true":
            return username, bearer_token

        # else exchange bearer token for PAT
        pat = self._exchange_bearer_for_pat(vsts_authority, bearer_token)
        return username, pat

    def get_credentials(self, url: str) -> tuple[str | None, str | None]:
        """
        Get the credentials for the given URL.

        Parameters
        ----------
        url : str
            The URL for which to get the credentials.

        Returns
        -------
        tuple[str | None, str | None]
            The username and password for the given URL.
        """
        # Public feed short circuit: return nothing if not getting credentials for the upload endpoint
        # (which always requires auth) and the endpoint is public (can authenticate without credentials).
        if not self._is_upload_endpoint(url) and self._can_authenticate(url, None):
            return None, None

        # Return personal access token if available
        if os.environ.get(self._PAT_ENV_VAR) and os.getenv(self._USE_BEARER_TOKEN_VAR_NAME, "False").lower() == "false":
            # Return the username and password from the environment variables
            return os.getenv(self._ADO_USERNAME_ENV_VAR, self._DEFAULT_USERNAME), os.environ.get(self._PAT_ENV_VAR)

        # Getting credentials; the credentials may come from the cache
        username, password = self._get_credentials_from_credential_provider(url)

        # Do not attempt to validate if the credentials could not be obtained
        if username is None or password is None:
            return username, password

        # Make sure the credentials are still valid (i.e. not expired)
        if self._can_authenticate(url, (username, password)):
            return username, password

        # Return None if the credentials are invalid
        return None, None



#!/usr/bin/env python3
"""
keyrings_artifacts/crpyt_file.py
---------------------------

This module provides an encrypted keyring backend.
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

import subprocess
from typing import TYPE_CHECKING

from keyrings.alt.file import EncryptedKeyring

_KEYGEN_SCRIPT = """
# Gather CPU information
cpu_id=$(lscpu | grep "Model name" | awk -F: '{print $2}' | xargs)

# Gather MAC address using ifconfig (first network interface)
mac_address_if=$(ifconfig | grep ether | head -n 1 | awk '{print $2}')

# Gather MAC address using ip (first network interface)
mac_address_ip=$(ip addr | grep ether | head -n 1 | awk '{print $2}')

# Gather hostname
hostname=$(hostname)

# Gather operating system name
uname_o=$(uname -o)

# Gather machine hardware name
uname_m=$(uname -m)

# Gather current username
whoami=$(whoami)

# Combine the information
combined_info="${cpu_id}_${mb_serial}_${disk_serial}_${mac_address_if}_${mac_address_ip}_${hostname}_${uname_o}_${uname_m}_${whoami}"

# Generate a unique hashed key using SHA-256
unique_key=$(echo -n "$combined_info" | sha256sum | awk '{print $1}')

# Output the unique key
echo "$unique_key"
"""


class EncryptedKeyring_(EncryptedKeyring):
    """
    Encrypted keyring backend.

    This backend provides a keyring that stores credentials in an encrypted file. The file is
    encrypted using the Fernet symmetric encryption algorithm. The encryption key is derived from
    a password generated from system and hardware information.
    """
    filename = "artifacts_keyring.cfg"
    priority = 3

    @property
    def _password(self):
        _command = f"bash -c '{_KEYGEN_SCRIPT}'"
        return subprocess.run(_command, shell=True, capture_output=True, check=False).stdout.decode("utf-8").strip()

    def _get_new_password(self):
        return self._password

    def _unlock(self):
        """
        Unlock this keyring by getting the password for the keyring from the
        user.
        """
        self.keyring_key = self._password
        try:
            ref_pw = self.get_password("keyring-setting", "password reference")
            if ref_pw != "password reference value":
                # keyring_key is generated from system and hardware information, so this can
                # only happen if hardware information has changed, which is unlikely. But if it
                # does happen, we need to initialize the keyring again.
                import pathlib

                try:
                    pathlib.Path(self.file_path).unlink()
                except FileNotFoundError:
                    pass
                self._init_file()
        except AssertionError:
            self._lock()
            raise ValueError("Incorrect Password")  # noqa: B904


if TYPE_CHECKING:
    EncryptedKeyring_()



#!/usr/bin/env python3
"""
keyrings_artifacts/support.py
---------------------------

It provides the AzureCredentialWithDevicecode class which extends ChainedTokenCredential to support
the following credential types:
    - AzureCliCredential
    - DefaultAzureCredential
    - ManagedIdentityCredential
    - SharedTokenCacheCredential
    - DeviceCodeCredential
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

import os
import webbrowser
from typing import Any

from azure.identity import (
    AzureCliCredential,
    ChainedTokenCredential,
    DeviceCodeCredential,
    EnvironmentCredential,
    InteractiveBrowserCredential,
    SharedTokenCacheCredential,
)

# Azure DevOps application default scope
# from https://github.com/microsoft/artifacts-credprovider/blob/cdc427e8236212b33041b4276961855b39bbe98d/CredentialProvider.Microsoft/CredentialProviders/Vsts/MSAL/MsalTokenProviderFactory.cs#L11
DEFAULT_SCOPE = "499b84ac-1321-427f-aa17-267ca6975798/.default"


class AzureCredentialWithDevicecode(ChainedTokenCredential):
    """
    A credential chain that is composed of multiple credentials.

    The following credential types are chained:
    - EnvironmentCredential
    - AzureCliCredential
    - SharedTokenCacheCredential
    - InteractiveBrowserCredential
    - DeviceCodeCredential

    Parameters
    ----------
    tenant_id : str, optional
        Tenant id to include in the token request.
    authority : str, optional
        Authority for the token request.
    additionally_allowed_tenants : list of str, optional
        Additional tenants to include in the token request.
    process_timeout : int, optional
        Timeout in seconds for the device code authentication process. Default is 90 seconds.
    scope : str, optional
        The scope for the token request. Default is "499b84ac-1321-427f-aa17-267ca6975798/.default".
        (From https://github.com/microsoft/artifacts-credprovider/blob/cdc427e8236212b33041b4276961855b39bbe98d/CredentialProvider.Microsoft/CredentialProviders/Vsts/MSAL/MsalTokenProviderFactory.cs#L11)
    with_az_cli : bool, optional
        Whether to include the Azure CLI credential in the chain. Default is True.
    """

    def __init__(  # noqa: PLR0913
        self,
        tenant_id: str = "",
        authority: str = "",
        additionally_allowed_tenants: list[str] | None = None,
        process_timeout: int = 180,
        scope: str = DEFAULT_SCOPE,
        with_az_cli: bool = True,
    ):
        self.scope = scope
        cred_chain = [
            *[
                # add environment credential if mandatory environment variable "AZURE_CLIENT_ID"
                # is set. EnvironmentCredential supports authenticating with service principal using
                # client secret, certificate, and managed identity. All three options require
                # AZURE_CLIENT_ID to be set.
                EnvironmentCredential() if os.getenv("AZURE_CLIENT_ID") else []
            ],
            *[
                AzureCliCredential(
                    tenant_id=tenant_id,
                    additionally_allowed_tenants=additionally_allowed_tenants,
                )
                # add Azure CLI credential if with_az_cli is True and tenant_id is set. If tenant_id
                # is not set, the Azure DevOps organization seems to a personal Microsoft account
                # without an Azure subscription, so the Azure CLI credential cannot be used.
                if with_az_cli and tenant_id
                else []
            ],
            SharedTokenCacheCredential(
                exclude_environment_credential=False,
                exclude_managed_identity_credential=True,
                authority=authority,
            ),
            *[
                InteractiveBrowserCredential(
                    tenant_id=tenant_id
                )
                # add interactive browser credential if a browser is available
                if self._is_interactive_browser_possible()
                else []
            ],
            DeviceCodeCredential(
                tenant_id=tenant_id,
                process_timeout=process_timeout,
            ),
        ]

        super().__init__(*(cred for cred in cred_chain if cred))

    def _is_interactive_browser_possible(self):
        """Check if the interactive browser credential is possible."""
        try:
            webbrowser.get()
            return True
        except webbrowser.Error:
            return False

    def get_token(self, *scopes: str, **kwargs):
        """
        Get an access token for the specified scopes.

        Parameters
        ----------
        scopes : str
            The scopes for the token request.
        kwargs : dict
            Additional keyword arguments.

        Returns
        -------
        AccessToken
            The access token.
        """
        if not scopes:
            scopes = (self.scope,)
        return super().get_token(*scopes, **kwargs)

    def __enter__(self) -> AzureCredentialWithDevicecode:
        return self

    def __exit__(self, *args: Any) -> None:
        pass



#!/usr/bin/env python3
"""
artifacts-keyring library
-------------------------

Keyring extension, that automatically retrieves credentials for Azure Artifacts.
"""

from __future__ import annotations

__author__ = "jslorrma"
__maintainer__ = "jslorrma"
__email__ = "jslorrma@gmail.com"

# version
try:
    from ._version import __version__, __version_tuple__, version
except ImportError:
    __version__ = version = "0.0.0"
    __version_tuple__ = (0, 0, 0)
